
////////////////////////////////////////////////////////////////////////////////
// uses

#include "flat_experience.h"
#include "flat_board.h"

#include <assert.h>
#include <algorithm>
#include <string.h>

#if (TARGET_DEVICE != TARGET_GP2XWIZ)
	enum
	{
		RGB565		= 0x4432,
		BGR565		= 0xC432,
		XRGB1555	= 0x4342,
		XBGR1555	= 0xC342,
		XRGB4444	= 0x4211,
		XBGR4444	= 0xC211,
		XRGB8332	= 0x4120,
		XBGR8332	= 0xC120,
		ARGB1555	= 0x3342,
		ABGR1555	= 0xB342,
		ARGB4444	= 0x2211,
		ABGR4444	= 0xA211,
		ARGB8332	= 0x1120,
		ABGR8332	= 0x9120,
		RGB888		= 0x4653,
		BGR888		= 0xC653,
		XRGB8888	= 0x4653,
		XBGR8888	= 0xC653,
		ARGB8888	= 0x0653,
		ABGR8888	= 0x8653,
		PTRGB565	= 0x443A
	};

	typedef unsigned short uint16_t;
	typedef unsigned long  uint32_t;

	extern int  lc_init(void);
	extern void lc_exit(void);
	extern void lc_setlayer(int layer, bool onoff, bool alpha, bool invert, bool trans, uint32_t mode);
	extern void lc_layerpos(int layer, int x1, int y1, int x2, int y2);
	extern void lc_settranscol(int layer, uint32_t colour);
	extern void lc_screensize(int w, int h);
	extern void lc_flipfb(int layer);
#else
	extern "C"
	{
		#include "castor.h"
	}
	#include <unistd.h>
#endif

////////////////////////////////////////////////////////////////////////////////
// external

extern volatile uint16_t *fb0_0, *fb0_1;
extern volatile uint16_t *fb1_0, *fb1_1;

////////////////////////////////////////////////////////////////////////////////
// using

using namespace flat_board;
using namespace flat_experience;

const unsigned short INTERNAL_COLOR_KEY      = 0xFFFE;
const unsigned short INTERNAL_REPLACED_COLOR = 0xFFFD;

////////////////////////////////////////////////////////////////////////////////
// Canvas implementation

#define MAX_SCREEN            2
#define MAX_BACK_BUFFER       2

#define USE_WIZ_CONFIG

#if defined(USE_WIZ_CONFIG)
	#define REVISE_COORD_X(canvas_info, x, y)  (y)
	#define REVISE_COORD_Y(canvas_info, x, y)  (canvas_info.height-(x))
#else
	#define REVISE_COORD_X(canvas_info, x, y)  (x)
	#define REVISE_COORD_Y(canvas_info, x, y)  (y)
//	#define REVISE_COORD_X(canvas_info, x, y)  (canvas_info.width-(y))
//	#define REVISE_COORD_Y(canvas_info, x, y)  (x)
#endif

class CCanvasNull: public ICanvas
{
public:
	CCanvasNull()
	{
		memset(&m_canvas_info, 0, sizeof(m_canvas_info));
		m_canvas_info.pixel_format  = PIXELFORMAT_INVALID;
	}
	virtual ~CCanvasNull(void) {}
	virtual  const TCanvasInfo& GetInfo(void) { return m_canvas_info; }
	virtual  void  BeginDraw(void) {}
	virtual  void  EndDraw(void) {}
	virtual  void  FillRect(int x, int y, int w, int h, unsigned long color_32bits, unsigned long alpha = 255) {}
	virtual  void  BitBlt(int x_dest, int y_dest, ICanvas* p_src_canvas, int x_sour, int y_sour, int w_sour, int h_sour, unsigned long alpha = 255) {}

private:
	TCanvasInfo    m_canvas_info;

};

template <TPixelFormat pixel_format>
class CWizCanvas: public ICanvas
{
protected:
	typedef CFlatBoard<pixel_format> TWizFlatBoard;
	typedef typename TPixelInfo<pixel_format>::TPixel TPixel;

public:
	CWizCanvas(TPixel* p_buffer, int width, int height, int pitch, bool has_ownership)
	:	m_has_ownership(has_ownership),
		m_p_buffer(p_buffer),
		m_flat_board0(p_buffer, width, height, pitch)
	{
		memset(&m_canvas_info, 0, sizeof(m_canvas_info));
		m_canvas_info.pixel_format  = pixel_format;
		m_canvas_info.width         = width;
		m_canvas_info.height        = height;
		m_canvas_info.depth         = sizeof(TPixel) * 8;
		m_canvas_info.has_color_key = false;

		m_p_flat_board = &m_flat_board0;
	}
	virtual ~CWizCanvas()
	{
		if (m_has_ownership)
			delete[] m_p_buffer;
	}
	virtual  const TCanvasInfo& GetInfo(void)
	{
		return m_canvas_info;
	}
	virtual  void  FillRect(int x, int y, int w, int h, unsigned long color_32bits, unsigned long alpha = 255)
	{
		if (w <= 0 || h <= 0)
			return;

		int x1 = REVISE_COORD_X(m_canvas_info, x, y);
		int y1 = REVISE_COORD_Y(m_canvas_info, x, y);
		int x2 = REVISE_COORD_X(m_canvas_info, x+w, y+h);
		int y2 = REVISE_COORD_Y(m_canvas_info, x+w, y+h);

		TPixel target_color = TPixelInfo<pixel_format>::ConvertColor(color_32bits);

#if defined(USE_WIZ_CONFIG)
		m_p_flat_board->FillRect(x1, y2, x2-x1, y1-y2, target_color, alpha);
#else
		m_p_flat_board->FillRect(x1, y1, x2-x1, y2-y1, target_color, alpha);
#endif
	}
	virtual  void  BitBlt(int x_dest, int y_dest, ICanvas* p_src_canvas, int x_sour, int y_sour, int w_sour, int h_sour, unsigned long alpha = 255)
	{
		if (p_src_canvas && p_src_canvas->GetInfo().pixel_format == pixel_format)
		{
			int x1_dest = REVISE_COORD_X(m_canvas_info, x_dest, y_dest);
			int y2_dest = REVISE_COORD_Y(m_canvas_info, x_dest+w_sour, y_dest+h_sour);

			int x1_sour = REVISE_COORD_X(p_src_canvas->GetInfo(), x_sour, y_sour);
			int y1_sour = REVISE_COORD_Y(p_src_canvas->GetInfo(), x_sour, y_sour);
			int x2_sour = REVISE_COORD_X(p_src_canvas->GetInfo(), x_sour+w_sour, y_sour+h_sour);
			int y2_sour = REVISE_COORD_Y(p_src_canvas->GetInfo(), x_sour+w_sour, y_sour+h_sour);

			TWizFlatBoard* flat_board = static_cast<CWizCanvas<pixel_format>*>(p_src_canvas)->m_GetFlatBoard();

#if defined(USE_WIZ_CONFIG)
			m_p_flat_board->BitBlt(x1_dest, y2_dest, flat_board, x1_sour, y2_sour, x2_sour-x1_sour, y1_sour-y2_sour, alpha);
#else
			m_p_flat_board->BitBlt(x1_dest, y2_dest, flat_board, x1_sour, y1_sour, x2_sour-x1_sour, y2_sour-y1_sour, alpha);
#endif
		}
	}

protected:
	bool           m_has_ownership;
	TPixel*        m_p_buffer;
	TWizFlatBoard  m_flat_board0;
	TWizFlatBoard* m_p_flat_board;
	TCanvasInfo    m_canvas_info;

	TWizFlatBoard* m_GetFlatBoard(void)
	{
		return m_p_flat_board;
	}
};

template <TPixelFormat pixel_format>
class CWizCanvasFlip: public CWizCanvas<pixel_format>
{
	typedef typename CWizCanvas<pixel_format>::TPixel        TPixel;
	typedef typename CWizCanvas<pixel_format>::TWizFlatBoard TWizFlatBoard;

public:
	CWizCanvasFlip(unsigned int screen_index, TPixel* p_buffer1, TPixel* p_buffer2, int width, int height, int pitch)
	:	CWizCanvas<pixel_format>(p_buffer1, width, height, pitch, false),
		m_screen_index(screen_index), m_width(width), m_height(height),
		m_flat_board1(p_buffer2, width, height, pitch)
	{
	}
	virtual  void  BeginDraw(void)
	{
		this->m_p_flat_board->FillRect(0, 0, m_width, m_height, 0);
	}
	virtual  void  EndDraw(void)
	{
#if 0
		lc_flipfb(m_screen_index);
		this->m_p_flat_board = (this->m_p_flat_board != &this->m_flat_board0) ? &this->m_flat_board0 : &this->m_flat_board1;
#else
	this->m_flat_board1.BitBlt(0, 0, &this->m_flat_board0, 0, 0, m_width, m_height);
#endif
	}

protected:
	unsigned int   m_screen_index;
	unsigned int   m_width;
	unsigned int   m_height;
	TWizFlatBoard  m_flat_board1;

};

template <TPixelFormat pixel_format>
class CWizCanvasUserAssigned: public CWizCanvas<pixel_format>
{
	typedef typename CWizCanvas<pixel_format>::TPixel TPixel;

public:
	CWizCanvasUserAssigned(unsigned int width, unsigned int height, void* p_buffer, bool use_color_key)
	:	CWizCanvas<pixel_format>(new TPixel[width * height], height, width, height, true)
	{
		TPixel* p_dest = this->m_p_buffer;

#if defined(USE_WIZ_CONFIG)
		for (int y = 0; y < int(width); y++)
		{
			TPixel* p_sour = (TPixel*)p_buffer + (width-1) - y;
			for (int x = 0; x < int(height); x++)
			{
				*p_dest++ = *p_sour;
				p_sour += width;
			}
		}
#else
		for (int y = 0; y < int(height); y++)
		{
			TPixel* p_sour = (TPixel*)p_buffer + width * y;
			for (int x = 0; x < int(width); x++)
			{
				*p_dest++ = *p_sour++;
			}
		}
#endif

		this->m_canvas_info.has_color_key = use_color_key;
		this->m_flat_board0.SetChromaKey(INTERNAL_COLOR_KEY);
	}
	virtual  void  BeginDraw(void) { }
	virtual  void  EndDraw(void) { }
};

template <typename TPixel>
static void s_ApplyTransparentColor(TPixel* p_buffer, unsigned int width, unsigned int height, unsigned long transparent_color)
{
	if ((p_buffer == NULL) && (transparent_color == INTERNAL_COLOR_KEY))
		return;

	TPixel* p_buffer_end = (TPixel*)p_buffer + width * height;

	--p_buffer;
	while (++p_buffer < p_buffer_end)
	{
		if (*p_buffer == INTERNAL_COLOR_KEY)
			*p_buffer = INTERNAL_REPLACED_COLOR;
		else if (*p_buffer == transparent_color)
			*p_buffer = INTERNAL_COLOR_KEY;
	}
}


////////////////////////////////////////////////////////////////////////////////
// namespace flat_experience

namespace flat_experience
{
	static ICanvas*    p_canvas[MAX_SCREEN];
	static CCanvasNull null_canvas;

	bool Initialize(TScreenRotation rotation)
	{
		typedef unsigned long*  PU32; 
		typedef unsigned short* PU16; 

		assert(MAX_SCREEN >= 2);

		if (lc_init() < 0)
			return false;
/*
		switch (format)
		{
		case SCREENFORMAT_ARGB8888_RGB565:
			lc_setlayer(0, true, true, false, false, ARGB8888);
			p_canvas[0] = new CWizCanvasFlip<PIXELFORMAT_ARGB8888>(0, PU32(fb0_0), PU32(fb0_1), 240, 320, 240);
			p_canvas[1] = new CWizCanvasFlip<PIXELFORMAT_RGB565>  (1, PU16(fb1_0), PU16(fb1_1), 240, 320, 240);
			break;
		case SCREENFORMAT_RGB565_RGB565:
			lc_setlayer(0, true, false, false, true, RGB565);
			lc_settranscol(0, 0);
			p_canvas[0] = new CWizCanvasFlip<PIXELFORMAT_RGB565>(0, PU16(fb0_0), PU16(fb0_1), 240, 320, 240);
			p_canvas[1] = new CWizCanvasFlip<PIXELFORMAT_RGB565>(1, PU16(fb1_0), PU16(fb1_1), 240, 320, 240);
			break;
		case SCREENFORMAT_DISABLE_RGB565:
			lc_setlayer(0, false, false, false, false, RGB565);
			p_canvas[0] = new CCanvasNull;
			p_canvas[1] = new CWizCanvasFlip<PIXELFORMAT_RGB565>(1, PU16(fb1_0), PU16(fb1_1), 240, 320, 240);
			break;
		default:
			return false;
		}
*/
		{
			lc_setlayer(0, false, false, false, false, RGB565);
			p_canvas[0] = new CCanvasNull;
			p_canvas[1] = new CWizCanvasFlip<PIXELFORMAT_RGB565>(1, PU16(fb1_0), PU16(fb1_1), 240, 320, 240);
		}		

		lc_layerpos(0, 0, 0, 239, 319);

		lc_setlayer(1, true, false, false, false, PTRGB565);
		lc_screensize(240, 320);
		lc_layerpos(1, 0, 0, 239, 319);

		return true;
	}

	void Finalize(void)
	{
		delete p_canvas[1];
		delete p_canvas[0];

		lc_screensize(320, 240);
		lc_layerpos(1, 0, 0, 319, 239);

		lc_exit();

	#if (TARGET_DEVICE != TARGET_GP2XWIZ)
	#else
		chdir("/usr/gp2x");
		execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
	#endif
	}

	ICanvas* GetCanvas(unsigned int screen_index)
	{
		if (screen_index >= MAX_SCREEN)
			return &null_canvas;

		return p_canvas[screen_index];
	}

	static ICanvas* s_CreateCanvas(TPixelFormat pixel_format, unsigned int width, unsigned int height, void* p_buffer, bool use_color_key)
	{
		if (width == 0 || height == 0 || p_buffer == 0)
			return &null_canvas;

		switch (pixel_format)
		{
		case PIXELFORMAT_RGB565:
			return new CWizCanvasUserAssigned<PIXELFORMAT_RGB565>(width, height, p_buffer, use_color_key);
		case PIXELFORMAT_ARGB8888:
			return new CWizCanvasUserAssigned<PIXELFORMAT_ARGB8888>(width, height, p_buffer, use_color_key);
		default:
			return &null_canvas;
		}
	}

	ICanvas* CreateCanvas(TPixelFormat pixel_format, unsigned int width, unsigned int height, void* p_buffer)
	{
		return s_CreateCanvas(pixel_format, width, height, p_buffer, false);
	}

	ICanvas* CreateCanvas(TPixelFormat pixel_format, unsigned int width, unsigned int height, void* p_buffer, unsigned long transparent_color)
	{
		if (width == 0 || height == 0 || p_buffer == 0)
			return &null_canvas;

		switch (pixel_format)
		{
		case PIXELFORMAT_RGB565:
			s_ApplyTransparentColor<unsigned short>((unsigned short*)p_buffer, width, height, transparent_color);
			break;
		case PIXELFORMAT_ARGB8888:
			s_ApplyTransparentColor<unsigned long>((unsigned long*)p_buffer, width, height, transparent_color);
			break;
		default:
			return &null_canvas;
		}

		return s_CreateCanvas(pixel_format, width, height, p_buffer, true);
	}
}
