
#include "image_decoder_png.h"
#include <png.h>
#include <assert.h>

#if defined(_WIN32)
	#pragma comment(lib, "libz_mtd.lib")
	#pragma comment(lib, "libpng_mtd.lib")
#endif

#define CHECK_FILE_BROKEN(_begin, _end, required_bytes) if (int(_end - _begin) < required_bytes) return false;

static void s_ReadDataFromBuffer(png_structp png_ptr, png_bytep data, png_size_t length)
{
	png_memcpy(data, png_ptr->io_ptr, length);
	png_ptr->io_ptr = (png_bytep)((unsigned int)png_ptr->io_ptr + length);
}

bool image_decoder::png::Decode(const unsigned char* p_stream, unsigned int stream_size, TImage& image)
{
	if (p_stream == 0)
		return false;

	if (stream_size == 0)
		return false;

	const unsigned char* p_stream_begin = p_stream;
	const unsigned char* p_stream_end   = p_stream + stream_size;

	CHECK_FILE_BROKEN(p_stream_begin, p_stream_end, 8);

	if (png_sig_cmp((png_byte*)p_stream_begin, 0, 8))
		return false;

	png_structp png_ptr  = NULL;
	png_infop   info_ptr = NULL;

	try
	{
		png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
		if (!png_ptr)
			throw 0;

		info_ptr = png_create_info_struct(png_ptr);
		if (!info_ptr)
			throw 0;

		png_set_read_fn(png_ptr, (png_voidp) p_stream_begin, s_ReadDataFromBuffer);

		png_read_info(png_ptr, info_ptr);

		bool is_acceptable = (info_ptr->color_type == PNG_COLOR_TYPE_RGB       && info_ptr->pixel_depth == 24) ||
							 (info_ptr->color_type == PNG_COLOR_TYPE_RGB_ALPHA && info_ptr->pixel_depth == 32) ||
							 (info_ptr->color_type == PNG_COLOR_TYPE_PALETTE   && info_ptr->pixel_depth ==  8);

		if (!is_acceptable)
			throw 1;

		{
			memset(&image, 0, sizeof(image));

			image.width          = info_ptr->width;
			image.height         = info_ptr->height;
			image.bytes_per_line = info_ptr->rowbytes;
			image.depth          = info_ptr->pixel_depth;
			image.p_bitmap       = new unsigned char[(sizeof(png_bytep) + info_ptr->rowbytes) * info_ptr->height];

			if (!image.p_bitmap)
				throw 2;

			if (info_ptr->color_type == PNG_COLOR_TYPE_PALETTE)
			{
				assert((1 << info_ptr->pixel_depth) >= info_ptr->num_palette);
				assert(info_ptr->num_palette <= 256);

				for (int i = 0; i < info_ptr->num_palette; i++)
				{
					image.palette[i] = 0xFF000000 | ((unsigned long)info_ptr->palette[i].red << 16) | ((unsigned long)info_ptr->palette[i].green << 8) | ((unsigned long)info_ptr->palette[i].blue);
				}
			}

			int number_of_passes = png_set_interlace_handling(png_ptr);
			png_read_update_info(png_ptr, info_ptr);

			{
				png_bytep  p_row = (png_bytep)image.p_bitmap;
				png_bytep* row_pointers = (png_bytep*)p_row;

				p_row += (sizeof(png_bytep) * info_ptr->height);

				for (int i = 0; i < int(info_ptr->height); i++)
				{
					row_pointers[i] = p_row;
					p_row += info_ptr->rowbytes;
				}

				png_read_image(png_ptr, row_pointers);
			}

			png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

			return true;
		}
	}
	catch (...)
	{
		if (png_ptr)
		{
			png_destroy_read_struct(&png_ptr, (info_ptr) ? &info_ptr : (png_infopp)NULL, (png_infopp)NULL);
		}
	}

	return false;
}
